#!/usr/bin/python

from __future__ import print_function
import os
import sys

class Register(object):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return 'BPF_REG_' + str(self.value)

class Immediate(object):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        if self.value > 2147483648:
            return '-' + str(4294967296 - self.value)
        return str(self.value)

    def equals(self, value):
        return self.value == value

class Offset(object):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        if self.value > 32768:
            return '-' + str(65536 - self.value)
        return str(self.value)

    def equals(self, value):
        return self.value == value

class InstructionOp(object):
    add = 0x00
    sub = 0x10
    mul = 0x20
    div = 0x30
    or_ = 0x40
    and_ = 0x50
    lsh = 0x60
    rsh = 0x70
    neg = 0x80
    mod = 0x90
    xor = 0xa0
    mov = 0xb0
    arsh = 0xc0
    end = 0xd0

    def __init__(self, value):
        self.value = value

    def __str__(self):
        if self.value == self.add:
            return 'BPF_ADD'
        elif self.value == self.sub:
            return 'BPF_SUB'
        elif self.value == self.mul:
            return 'BPF_MUL'
        elif self.value == self.mul:
            return 'BPF_MUL'
        elif self.value == self.div:
            return 'BPF_DIV'
        elif self.value == self.or_:
            return 'BPF_OR'
        elif self.value == self.and_:
            return 'BPF_AND'
        elif self.value == self.lsh:
            return 'BPF_LSH'
        elif self.value == self.rsh:
            return 'BPF_RSH'
        elif self.value == self.neg:
            return 'BPF_NEG'
        elif self.value == self.mod:
            return 'BPF_MOD'
        elif self.value == self.xor:
            return 'BPF_XOR'
        elif self.value == self.mov:
            return 'BPF_MOV'
        elif self.value == self.arsh:
            return 'BPF_ARSH'
        elif self.value == self.end:
            return 'BPF_END'
        else:
            return 'UNKNOWN OP %d' % (self.value)

    def equals(self, value):
        return self.value == value

class JMPInstructionOp(object):
    ja = 0x00
    jeq = 0x10
    jgt = 0x20
    jge = 0x30
    jset = 0x40
    jne = 0x50
    jsgt = 0x60
    jsge = 0x70
    call = 0x80
    ext = 0x90

    def __init__(self, value):
        self.value = value

    def __str__(self):
        if self.value == self.ja:
            return 'BPF_JA'
        elif self.value == self.jeq:
            return 'BPF_JEQ'
        elif self.value == self.jgt:
            return 'BPF_JGT'
        elif self.value == self.jge:
            return 'BPF_JGE'
        elif self.value == self.jset:
            return 'BPF_JSET'
        elif self.value == self.jne:
            return 'BPF_JNE'
        elif self.value == self.jsgt:
            return 'BPF_JSGT'
        elif self.value == self.jsge:
            return 'BPF_CALL'
        elif self.value == self.ext:
            return 'BPF_EXIT'

    def equals(self, value):
        return self.value == value

class InstructionMode(object):
    imm = 0x00
    abso = 0x20
    ind = 0x40
    mem = 0x60
    leng = 0x80
    msh = 0xa0
    xadd = 0xc0

class InstructionSource(object):
    k = 0x00
    x = 0x08

class InstructionSize(object):
    word = 0x00
    half = 0x08
    byte = 0x10
    double = 0x18

    def __init__(self, value):
        self.value = value

    def __str__(self):
        if self.value == self.word:
            return 'BPF_W'
        elif self.value == self.half:
            return 'BPF_H'
        elif self.value == self.byte:
            return 'BPF_B'
        else:
            return 'BPF_DW'
        return 'WTF'

    def equals(self, value):
        return self.value == value

class InstructionClass(object):
    ld = 0x00
    ldx = 0x01
    st = 0x02
    stx = 0x03
    alu = 0x04
    jmp = 0x05
    ret = 0x06
    alu64 = 0x07

    def __init__(self, op, value, source, mode, size, dst, src, imm, off):
        self.op = InstructionOp(op)
        self.value = value
        self.source = source
        self.mode = mode
        self.size = InstructionSize(size)
        self.dst = Register(dst)
        self.src = Register(src)
        self.imm = Immediate(imm)
        self.off = Offset(off)

    def needNext(self):
        return False

class InstructionClassLD(InstructionClass):

    def __str__(self):
        if self.mode == InstructionMode.imm:
            if self.size.equals(InstructionSize.double):
                if self.src.value == 1:
                    return 'BPF_LD_MAP_FD(' + str(self.dst) + ',' + str(self.imm) + ')'
                else:
                    # for this one, we bypass the imm printing code
                    return 'BPF_LD_IMM64_RAW(' + str(self.dst) + ',' + str(self.src) + ',' + str(self.imm.value) + ')'
                    #return 'BPF_LD_IMM64_RAW(' + str(self.dst) + ',' + str(self.src) + ',' + str(self.imm) + ')'
            else:
                return 'BPF_RAW_INSN(BPF_LD|BPF_IMM,' + str(self.dst) + ',' + str(self.src) + ',' + str(self.off) + ',' + str(self.imm) + ')'
        elif self.mode == InstructionMode.abso:
            return 'BPF_LD_ABS(' + str(self.size) + ',' + str(self.imm) + ')'
        elif self.mode == InstructionMode.ind:
            return 'BPF_RAW_INSN(BPF_LD|BPF_IND|' + str(self.size) + ',' + str(self.dst) + ',' + str(self.src) + ',' + str(self.off) + ',' + str(self.imm) + ')'
        else:
            raise Exception('Unknown LD mode %d' % (self.mode))

    def needNext(self):
        if self.mode == InstructionMode.imm and self.size.equals(InstructionSize.double):
            return True
        return False

    def feedNext(self, value):
        if self.mode == InstructionMode.imm and self.size.equals(InstructionSize.double):
            self.imm.value = self.imm.value + (value << 32)
        else:
            raise Exception('Unneeded next value')

class InstructionClassLDX(InstructionClass):

    def __str__(self):
        return 'BPF_LDX_MEM(' + str(self.size) + ',' + str(self.dst) + ',' + str(self.src) + ',' + str(self.off) + ')'

class InstructionClassST(InstructionClass):

    def __str__(self):
        return 'BPF_ST_MEM(' + str(self.size) + ',' + str(self.dst) + ',' + str(self.src) + ',' + str(self.off) + ')'

class InstructionClassSTX(InstructionClass):

    def __str__(self):
        if self.mode == InstructionMode.mem:
            return 'BPF_STX_MEM(' + str(self.size) + ',' + str(self.dst) + ',' + str(self.src) + ',' + str(self.off) + ')'
        elif self.mode == InstructionMode.xadd:
            return 'BPF_RAW_INSN(BPF_STX|BPF_XADD|' + str(self.size) + ',' + str(self.dst) + ',' + str(self.src) + ',' + str(self.off) + ',' + str(self.imm) + ')'
        else:
            raise Exception('Unknown STX mode %d' % (self.mode))

class InstructionClassALU(InstructionClass):

    def __str__(self):
        if self.source == InstructionSource.k:
            return 'BPF_ALU32_IMM(' + str(self.op) + ',' + str(self.dst) + ',' + str(self.imm) + ')'
        else:
            return 'BPF_ALU32_REG(' + str(self.op) + ',' + str(self.dst) + ',' + str(self.src) + ')'

class InstructionClassJMP(InstructionClass):

    def __init__(self, op, value, source, mode, size, dst, src, imm, off):
        self.op = JMPInstructionOp(op)
        self.value = value
        self.source = source
        self.mode = mode
        self.size = InstructionSize(size)
        self.dst = Register(dst)
        self.src = Register(src)
        self.imm = Immediate(imm)
        self.off = Offset(off)

    def immToFunc(self):
        if self.imm.value == 1:
            return 'BPF_FUNC_map_lookup_elem'
        elif self.imm.value == 3:
            return 'BPF_FUNC_map_update_elem'
        elif self.imm.value == 4:
            return 'BPF_FUNC_map_delete_elem'
        elif self.imm.value == 6:
            return 'BPF_FUNC_trace_printk'
        elif self.imm.value == 12:
            return 'BPF_FUNC_tail_call'
        elif self.imm.value == 26:
            return 'BPF_FUNC_skb_load_bytes'
        else:
            raise Exception('Unknown function (%s)' % str(self.imm.value))

    def __str__(self):
        if self.op.value == JMPInstructionOp.ext:
            return 'BPF_EXIT_INSN()'
        elif self.op.value == JMPInstructionOp.call:
            return 'BPF_RAW_INSN(BPF_JMP|BPF_CALL,0,0,0,' + self.immToFunc() + ')'
        if self.source == InstructionSource.x:
            return 'BPF_JMP_REG(' + str(self.op) + ',' + str(self.dst) + ',' + str(self.src) + ',' + str(self.off) + ')'
        else:
            return 'BPF_JMP_IMM(' + str(self.op) + ',' + str(self.dst) + ',' + str(self.imm) + ',' + str(self.off) + ')'

class InstructionClassRET(InstructionClass):

    def __str__(self):
        return 'BUG'

class InstructionClassALU64(InstructionClass):

    def __str__(self):
        if self.op.value == InstructionOp.mov:
            if self.source == InstructionSource.x:
                return 'BPF_MOV64_REG(' + str(self.dst) + ',' + str(self.src) + ')'
            else:
                return 'BPF_MOV64_IMM(' + str(self.dst) + ',' + str(self.imm) + ')'

        if self.source == InstructionSource.x:
            return 'BPF_ALU64_REG(' + str(self.op) + ',' + str(self.dst) + ',' + str(self.src) + ')'
        else:
            return 'BPF_ALU64_IMM(' + str(self.op) + ',' + str(self.dst) + ',' + str(self.imm) + ')'


def getInstClass(inst, dst, src, off, imm):
    op = inst & 0xf0
    value = inst & 0x07
    source = inst & 0x08
    size = inst & 0x18
    mode = inst & 0xe0

    if value == InstructionClass.ld:
        return InstructionClassLD(op, value, source, mode, size, dst, src, imm, off)
    elif value == InstructionClass.ldx:
        return InstructionClassLDX(op, value, source, mode, size, dst, src, imm, off)
    elif value == InstructionClass.st:
        return InstructionClassST(op, value, source, mode, size, dst, src, imm, off)
    elif value == InstructionClass.stx:
        return InstructionClassSTX(op, value, source, mode, size, dst, src, imm, off)
    elif value == InstructionClass.alu:
        return InstructionClassALU(op, value, source, mode, size, dst, src, imm, off)
    elif value == InstructionClass.jmp:
        return InstructionClassJMP(op, value, source, mode, size, dst, src, imm, off)
    elif value == InstructionClass.ret:
        return InstructionClassRET(op, value, source, mode, size, dst, src, imm, off)
    elif value == InstructionClass.alu64:
        return InstructionClassALU64(op, value, source, mode, size, dst, src, imm, off)

class Instruction(object):

    def __init__(self, value):
        self._class = getInstClass(value)

    def __str__(self):
        return str(self._class)


def reverse(content):
    size = len(content)
    if (size % 8) != 0:
        sys.exit('Content length is not a multiple of 8, something is wrong.')

    count = size / 8
    idx = 0
    while idx < count:
        trad = ''
        inst = content[idx*8]
        dst = content[(idx*8)+1] % 16
        src = int(content[(idx*8)+1] / 16)

        off = (content[(idx*8)+3]*256) + content[(idx*8)+2]
        imm = (content[(idx*8)+7]*16777216) + (content[(idx*8)+6]*65536) + (content[(idx*8)+5]*256) + content[(idx*8)+4]
        value = getInstClass(inst, dst, src, off, imm)
        parsedIdx = idx
        if value.needNext():
            idx = idx + 1
            imm = (content[(idx*8)+7]*16777216) + (content[(idx*8)+6]*65536) + (content[(idx*8)+5]*256) + content[(idx*8)+4]
            value.feedNext(imm)

        print('%s,' % (value))
#        print('%d: %s' % (parsedIdx, value))
#        print('%d: %s: %s' % (parsedIdx, ":".join("{:02x}".format(c) for c in content[parsedIdx*8:(parsedIdx+1)*8]), value))

        idx = idx + 1

if __name__ == "__main__":
    if len(sys.argv) != 2:
        sys.exit('Usage: %s <eBPF binary file>' % (sys.argv[0]))

    if not os.path.isfile(sys.argv[1]):
        sys.exit("eBPF binary file %s does not exist!" % (sys.argv[1]))

    with open(sys.argv[1], 'rb') as bpfFile:
        reverse(bpfFile.read())
